
#include "matrix.h"
#include <stdio.h>
#include <string.h>
#include <stdbool.h>
#include <stdlib.h>
#include <time.h>
#include <signal.h>
#include <unistd.h>
#include <sys/time.h>

#ifdef GFX_SDL
#include <SDL/SDL.h>
#endif

void run(int w, int h, unsigned int cycles);

struct config {
	char c;
	bool buffered : 1;
	bool vt100 : 1;
	bool wait : 1;
	bool quiet: 1;
#ifdef GFX_SDL
	bool gfx : 1;
#endif
	unsigned short sleep;
};
static struct config cfg = {
	'#',
	false, true, true, false,
#ifdef GFX_SDL
	false,
#endif
	(unsigned short)0
};

const static char *call;

void usage(int code) {
	printf("%s [OPTIONS] [COUNT]\n", call);
	puts("  -l NUM   number of lines");
	puts("  -c NUM   number of columns");
	puts("  -b       use buffered console output");
	puts("  -n       don't wait for user input between generations");
	puts("  -D       don't use VT100 escape codes for console display");
	puts("  -C CHAR  use CHAR as marker for alive entities");
	puts("  -q       quiet mode, no console output");
	puts("  -s TIME  wait TIME milliseconds between generations");
#ifdef GFX_SDL
	puts("  -g       enable graphic mode, displays a window");
	puts("  -f       enable fullscreen display");
#endif
	puts("  COUNT    number of generations to calculate");
	puts("");
	puts("This program is Free Software and released under the terms of the");
	puts("General Public License v2 by Franz Brausse");
	puts("");
	puts("Hint: The shell command \"export LINES COLUMNS\" will enable automatic field size");
	puts("      adjustment to fit the console");
	exit(code);
}

static const char t_cls[] = { 0x1b, 0x5b, '2', 'J', 0x00 };		// ^[[2J - clear screen
static const char t_ulc[] = { 0x1b, 0x5b, 'f', 0x00 };			// ^[[f  - move cursor to upper left
static const char t_clr[] = { 0x1b, 0x5b, 'K', 0x00 };			// ^[[K  - clear line to right

void handler(int signal);

static char *buf;

#ifdef GFX_SDL
static SDL_Surface *display;
static SDL_Surface *sf;
static SDL_Event events[10];
#endif

int main(int argc, const char *const *argv, const char *const *envp) {
	
	call = *argv;
	
	unsigned short c = 80;
	unsigned short l = 24;
	unsigned int count = 0xFFFFFFFF;
	
	const char *env_w = NULL, *env_h = NULL;
#ifdef GFX_SDL
	bool fs = false;
#endif
	
	const char *arg;
	while (!(env_h && env_w) && (arg = *envp++)) {
		if (!env_h && strstr(arg, "LINES=") == arg) {
			env_h = arg + strlen("LINES=");
		} else if (!env_w && strstr(arg, "COLUMNS=") == arg) {
			env_w = arg + strlen("COLUMNS=");
		}
	}
	/*
	env_w = getenv("COLUMNS");
	env_h = getenv("LINES");
	*/
	if (env_w)
		c = atoi(env_w);
	if (env_h)
		l = atoi(env_h);
	l--;
	
	// loop through arguments
	while ((arg = *++argv)) {
		if (strstr("-l", arg)) {			// lines
			if (!*++argv)
				usage(EXIT_FAILURE);
			if (!sscanf(*argv, "%hu", &l))
				usage(EXIT_FAILURE);
		} else if (strstr("-c", arg)) {		// columns
			if (!*++argv)
				usage(EXIT_FAILURE);
			if (!sscanf(*argv, "%hu", &c))
				usage(EXIT_FAILURE);
		} else if (strstr("-b", arg)) {		// buffered output
			cfg.buffered = true;
		} else if (strstr("-n", arg)) {		// don't wait for user input
			cfg.wait = false;
		} else if (strstr("-D", arg)) {		// non VT100 compatible terminal
			cfg.vt100 = false;
		} else if (strstr("-C", arg)) {		// character to draw
			if (!*++argv)
				usage(EXIT_FAILURE);
			cfg.c = **argv;
		} else if (strstr("-q", arg)) {		// quiet mode
			cfg.quiet = true;
		} else if (strstr("-s", arg)) {		// sleep in ms
			if (!*++argv)
				usage(EXIT_FAILURE);
			if (!sscanf(*argv, "%hu", &(cfg.sleep)))
				usage(EXIT_FAILURE);
		} else if (strstr("-h", arg)) {		// help
			usage(EXIT_SUCCESS);
#ifdef GFX_SDL
		} else if (strstr("-g", arg)) {		// graphics using SDL
			cfg.gfx = true;
		} else if (strstr("-f", arg)) {		// fullscreen mode
			fs = true;
#endif
		} else {							// number of generations
			if (!sscanf(arg, "%u", &count))
				usage(EXIT_FAILURE);
		}
	}
	
	if (false) {
		printf("lines: %u\n", l);
		printf("cols : %u\n", c);
		printf("buff : %d\n", cfg.buffered);
		printf("wait : %d\n", cfg.wait);
		printf("vt100: %d\n", cfg.vt100);
		printf("char : %c\n", cfg.c);
		printf("count: %u\n", count);
	}
	
	void *org_signal_handler = signal(SIGINT, handler);
	
	if (cfg.buffered)
		buf = (char *)malloc(l * (c + 1));
	if (!cfg.quiet && cfg.vt100)
		fputs(t_cls, stdout);
	
#ifdef GFX_SDL
	if (cfg.gfx) {
		SDL_Init(SDL_INIT_VIDEO | SDL_INIT_NOPARACHUTE | SDL_INIT_EVENTTHREAD);
		SDL_WM_SetCaption("Game of Life (ESC or Q to exit)", "Game of Life");
		int flags = SDL_HWSURFACE;
		if (fs) {
			SDL_ShowCursor(SDL_DISABLE);
			flags |= SDL_FULLSCREEN;
		}
		display = SDL_SetVideoMode(c, l, 0, flags);
	}
#endif
	
	run(c, l, count);
	
#ifdef GFX_SDL
	if (cfg.gfx) {
		if (fs)
			SDL_ShowCursor(SDL_ENABLE);
		SDL_Quit();
	}
#endif
	
	if (cfg.buffered)
		free(buf);
	
	signal(SIGINT, org_signal_handler);
	
	return EXIT_SUCCESS;
}

int print_matrix_direct(matrix *m) {
	int x, y, c = 0;
	for (y = 0; y < mtrx_height(m); y++) {
		for (x = 0; x < mtrx_width(m); x++)
			if (is_set(m, y, x)) {
				c++;
				putchar(cfg.c);
			} else {
				putchar(' ');
			}
		puts("");
	}
	return c;
}

int print_matrix_buffered(matrix *m) {
	int x, y, c = 0, i = 0;
	int h = mtrx_height(m), w = mtrx_width(m);
	
	for (y = 0; y < h; y++, i++) {
		for (x = 0; x < w; x++, i++)
			buf[i] = (is_set(m, y, x)) ? c++, cfg.c : ' ';
		buf[i] = '\n';
	}
	fputs(buf, stdout);
	return c;
}

void print_matrix(matrix *m) {
	int c;
	
	if (cfg.vt100)
		fputs(t_ulc, stdout);
	
	c = (cfg.buffered) ? print_matrix_buffered(m) : print_matrix_direct(m);
	
	if (cfg.vt100)
		fputs(t_clr, stdout);
	printf("alive: %d", c);
	if (!cfg.vt100)
		puts("");
	fflush(stdout);
}

bool interrupted = false;

inline void output(matrix *m) {
	if (!cfg.quiet) {
		print_matrix(m);
	}
#ifdef GFX_SDL
	if (cfg.gfx) {
		SDL_PumpEvents();
		int ne, i;
		while ((ne = SDL_PeepEvents(events, 10, SDL_GETEVENT, SDL_EVENTMASK(SDL_KEYDOWN))) > 0) {
			for (i=0; i<ne; i++) {
				switch (events[i].key.keysym.sym) {
					case SDLK_ESCAPE:
					case SDLK_q:
						interrupted = true;
						goto exit;
					default:
						break;
				}
			}
		}
		
		exit:
		SDL_BlitSurface(sf, NULL, display, NULL);
		SDL_Flip(display);
	}
	
#endif
}

#define change_cell(to, y, x, num)				\
		switch (num) {							\
			case 3: set(to, y, x);				\
			case 2: break;						\
			default: unset(to, y, x); break;	\
		}

void step(const matrix *, matrix *);

#if STEP==2
static char *tt;

void step(const matrix *from, matrix *to) {
	
	const unsigned int w = mtrx_width(from);
	const unsigned int h = mtrx_height(from);
	const int maxh = h - 1;
	const int maxw = w - 1;
	char *t = tt;
	memset(t, 0, w * h);
	int y, xmm, xpp, x;
	
	int yoffmm = (maxh - 1) * w, yoffpp;
	int yoff = maxh * w;
	
	if (is_set(from, 0, 0)) {
		t[    1]++;
		t[w    ]++;
		t[w + 1]++;
	}
	if (is_set(from, 0, maxw)) {
		t[    maxw - 1]++;
		t[w + maxw - 1]++;
		t[w + maxw    ]++;
	}
	if (is_set(from, maxh, 0)) {
		t[yoffmm    ]++;
		t[yoffmm + 1]++;
		t[yoff   + 1]++;
	}
	if (is_set(from, maxh, maxw)) {
		t[yoff   + maxw - 1]++;
		t[yoffmm + maxw - 1]++;
		t[yoffmm + maxw    ]++;
	}
	
	// upper and lower line
	for (x = 1; x < maxw; x++) {
		if (is_set(from, 0, x)) {
			t[    x - 1]++;
			t[    x + 1]++;
			t[w + x - 1]++;
			t[w + x    ]++;
			t[w + x + 1]++;
		}
		
		if (is_set(from, maxh, x)) {
			t[yoff   + x - 1]++;
			t[yoff   + x + 1]++;
			t[yoffmm + x - 1]++;
			t[yoffmm + x    ]++;
			t[yoffmm + x + 1]++;
		}
	}
	
	// left and right row
	yoff = w;
	for (y = 1; y < maxh; y++, yoff += w) {
		if (is_set(from, y, 0)) {
			t[yoff - w]++;
			t[yoff + w]++;
			t[yoff - w + 1]++;
			t[yoff     + 1]++;
			t[yoff + w + 1]++;
		}
		if (is_set(from, y, maxw)) {
			t[yoff - w + maxw]++;
			t[yoff + w + maxw]++;
			t[yoff - w + maxw - 1]++;
			t[yoff     + maxw - 1]++;
			t[yoff + w + maxw - 1]++;
		}
	}
	
	// center
	yoff = w, yoffmm = 0, yoffpp = 2 * w;
	for (y = 1; y < maxh; y++, yoff += w, yoffmm += w, yoffpp += w) {
		for (xmm = 0, x = 1, xpp = 2; x < maxw; x++, xmm++, xpp++) {
			if (is_set(from, y, x)) {
				t[yoffmm + xmm]++;
				t[yoffmm + x  ]++;
				t[yoffmm + xpp]++;
				t[yoff   + xmm]++;
				t[yoff   + xpp]++;
				t[yoffpp + xmm]++;
				t[yoffpp + x  ]++;
				t[yoffpp + xpp]++;
			}
	//		change_cell(to, ymm, x, (int)t[yoffmm + x]);
		}
	//	change_cell(to, ymm, 0, t[yoffmm]);			// left cell
	//	change_cell(to, ymm, x, t[yoffmm + maxw]);	// right cell
	}
	/*
	for (x = 0; x < w; x++) {
	//	change_cell(to, maxh - 1, x, t[yoffmm - w + x]);	// fore-last line
		change_cell(to, maxh    , x, t[yoffmm     + x]);	// lower line
	}*/
	
	// change cells accordingly
	yoff = 0;
	for (y = 0; y < h; y++, yoff += w) {
		for (x = 0; x < w; x++) {
	//		printf("%d ", t[yoff + x]);
			change_cell(to, y, x, t[yoff + x]);
		}
	//	puts("");
	}
}

#elif STEP==1

void step(register const matrix *from, matrix *to) {
	
	int x, y, num;
	const int maxw = mtrx_width(from) - 1;
	const int maxh = mtrx_height(from) - 1;
	const int maxhmm = maxh - 1;
	const int maxwmm = maxw - 1;
	
	// upper left corner
	num = (0
		+ is_set(from, 0, 1)
		+ is_set(from, 1, 0)
		+ is_set(from, 1, 1));
	change_cell(to, 0, 0, num);
	
	// upper right corner
	num = (0
		+ is_set(from, 0, maxwmm)
		+ is_set(from, 1, maxwmm)
		+ is_set(from, 1, maxw  ));
	change_cell(to, 0, maxw, num);
	
	// lower left corner
	num = (0
		+ is_set(from, maxhmm, 0)
		+ is_set(from, maxhmm, 1)
		+ is_set(from, maxh  , 1));
	change_cell(to, maxh, 0, num);
 	
	// lower right corner
	num = (0
		+ is_set(from, maxh  , maxwmm)
		+ is_set(from, maxhmm, maxwmm)
		+ is_set(from, maxhmm, maxw  ));
	change_cell(to, maxh, maxw, num);
	
	// upper border
	for (x = 1; x < maxw; x++) {
		const int xmm = x - 1, xpp = x + 1;
		num = (0
			+ is_set(from, 0, xmm)
			+ is_set(from, 0, xpp)
			+ is_set(from, 1, xmm)
			+ is_set(from, 1, x  )
			+ is_set(from, 1, xpp));
		change_cell(to, 0, x, num);
	}
	
	// lower border
	for (x = 1; x < maxw; x++) {
		const int xmm = x - 1, xpp = x + 1;
		num = (0
			+ is_set(from, maxh  , xmm)
			+ is_set(from, maxh  , xpp)
			+ is_set(from, maxhmm, xmm)
			+ is_set(from, maxhmm, x  )
			+ is_set(from, maxhmm, xpp));
		change_cell(to, maxh, x, num);
	}
 	
	// left border
	for (y = 1; y < maxh; y++) {
		const int ymm = y - 1, ypp = y + 1;
		num = (0
			+ is_set(from, ymm, 0)
			+ is_set(from, ypp, 0)
			+ is_set(from, ymm, 1)
			+ is_set(from, y  , 1)
			+ is_set(from, ypp, 1));
		change_cell(to, y, 0, num);
	}
	
	// right border
	for (y = 1; y < maxh; y++) {
		const int ymm = y - 1, ypp = y + 1;
		num = (0
			+ is_set(from, ymm, maxw  )
			+ is_set(from, ypp, maxw  )
			+ is_set(from, ymm, maxwmm)
			+ is_set(from, y  , maxwmm)
			+ is_set(from, ypp, maxwmm));
		change_cell(to, y, maxw, num);
	}
 	
	// center
	int ymm = 0, ypp = 2;
	for (y = 1; y < maxh; y++, ymm++, ypp++) {
		int xmm = 0, xpp = 2;
		for (x = 1; x < maxw; x++, xmm++, xpp++) {
			num = (0
				+ is_set(from, ymm, xmm)
				+ is_set(from, ymm, x  )
				+ is_set(from, ymm, xpp)
				+ is_set(from, y  , xmm)
				+ is_set(from, y  , xpp)
				+ is_set(from, ypp, xmm)
				+ is_set(from, ypp, x  )
				+ is_set(from, ypp, xpp));
			change_cell(to, y, x, num);
		}
 	}
}
#else
	#error "No algorithm for next generation defined, use STEP=1 or STEP=2"
#endif

void handler(int signal) {
	if (signal == SIGINT) {
		interrupted = true;
	} else {
		psignal(signal, "Unbekanntes Signal: ");
	}
}

static struct timeval _ts, _te;
static struct timeval *ts = &_ts, *te = &_te;

#define delta(a, b) ((b.tv_sec - a.tv_sec) * 1000000 + (b.tv_usec - a.tv_usec))

inline int measure2(void (*func)(const matrix *m1, matrix *m2), const matrix *m1, matrix *m2) {
	gettimeofday(ts, NULL);
	func(m1, m2);
	gettimeofday(te, NULL);
	return delta(_ts, _te);
}

inline int measure1(void (*func)(matrix *m), matrix *m) {
	gettimeofday(ts, NULL);
	func(m);
	gettimeofday(te, NULL);
	return delta(_ts, _te);
}

void run(int w, int h, unsigned int cycles) {
	matrix *m1, *m2;
	unsigned int run = 0;
	
	struct timeval _trs, _tre;
	
	mtrx_init(&m1, w, h);
	mtrx_init(&m2, w, h);
	mtrx_fill(m1);
	mtrx_clone_data(m1, m2);
	
#ifdef GFX_SDL
	sf = create_sdl_surface(m1);
#endif
	
#if STEP==2
	tt = (char *)malloc(w * h);
#endif
	
	int us_step = 0, us_clone = 0, us_print = 0;
	
	gettimeofday(&_trs, NULL);
	
	while (!interrupted) {
		us_print += measure1(output, m1);
		
		if (run++ == cycles)
			break;
		if (cfg.wait) {
			if (interrupted)
				break;
			getchar();
			if (interrupted)
				break;
		} else if (cfg.sleep > 0) {
			if (interrupted)
				break;
			usleep(1000UL * cfg.sleep);
			if (interrupted)
				break;
		}
		
		us_step += measure2(step, m1, m2);
		us_clone += measure2(mtrx_clone_data, m2, m1);
	}
	
	gettimeofday(&_tre, NULL);
	
	printf("\r");
	printf("print: %d ms\n", (us_print / 1000));
	printf("step : %d ms\n", (us_step / 1000));
	printf("clone: %d ms\n", (us_clone / 1000));
	const long int ms = delta(_trs, _tre) / 1000;
	printf("computed %d generations, %ld/sec\n", run, (ms > 0) ? run * 1000 / ms : 0L);
	
#if STEP==2
	free(tt);
#endif
	
#ifdef GFX_SDL
	SDL_FreeSurface(sf);
#endif
	
	mtrx_free(&m1);
	mtrx_free(&m2);
}
