
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>

#ifdef GFX_SDL
#include <SDL/SDL.h>
#endif

#include "matrix.h"

#ifndef TYPE
#	define TYPE char
#endif

// from http://graphics.stanford.edu/~seander/bithacks.html#IntegerLogDeBruijn
static const int MultiplyDeBruijnBitPosition[32] = {
	0, 1, 28, 2, 29, 14, 24, 3, 30, 22, 20, 15, 25, 17, 4, 8, 
	31, 27, 13, 23, 21, 19, 16, 7, 26, 12, 18, 6, 11, 5, 10, 9
};

const int lds = (sizeof(TYPE) * 0x077CB531UL) >> 27;

#define BITS (sizeof(TYPE) << 3)
#define LD_S (MultiplyDeBruijnBitPosition[lds] + 3)
#define MASK (BITS - 1)

struct matrix {
	unsigned TYPE * data;
	unsigned int height;
	unsigned int width, pwidth;
};

#ifdef GFX_SDL
SDL_Surface *create_sdl_surface(matrix *m) {
	return SDL_CreateRGBSurfaceFrom(m->data, m->width, m->height, 1, m->pwidth, 1, 0, 0, 1);
}
#endif

inline unsigned int mtrx_width(const matrix * m) {
	return m->width;
}

inline unsigned int mtrx_height(const matrix * m) {
	return m->height;
}

#if SDL_BYTEORDER == SDL_BIG_ENDIAN
#define mask(x) (x & MASK)
#else
#define mask(x) (MASK - (x & MASK))
#endif

bool is_set(const matrix *m, unsigned int y, unsigned int x) {
	const int idx = y * m->pwidth + (x >> LD_S);
#ifdef DEBUG
	printf("accessing idx (%d / %d) for (%d,%d): %08x @ %08x\n", idx, x & MASK, y, x, m->data[idx], (uint)m->data + (idx) * sizeof(TYPE));
#endif
	return (m->data[idx] & (1 << mask(x))) != 0;
}

void set(matrix *m, unsigned int y, unsigned int x) {
	int idx = y * m->pwidth + (x >> LD_S);
#ifdef DEBUG
	printf("setting   idx (%d / %d) for (%d,%d)\n", idx, x & MASK, y, x);
#endif
	m->data[idx] |= 1 << mask(x);
}

void unset(matrix *m, unsigned int y, unsigned int x) {
	int idx = y * m->pwidth + (x >> LD_S);
#ifdef DEBUG
	printf("unsetting idx (%d / %d) for (%d,%d): %08x, %08x @ %08x\n", idx, x & MASK, y, x, m->data[idx], ~((TYPE)1 << (x & MASK)), (uint)m->data + (idx) * sizeof(TYPE));
#endif
	m->data[idx] &= ~(1 << mask(x));
#ifdef DEBUG
	printf("afterwards: %08x\n", m->data[idx]);
#endif
}





inline bool _is_set(matrix * m, unsigned int y, unsigned int x) {
	int idx = y * m->width + x;
#ifdef DEBUG
	printf("accessing idx %d (%d / %d) for (%d,%d)\n", idx, idx >> LD_S, idx & MASK, y, x);
#endif
	return (m->data[idx >> LD_S] & (1 << (idx & MASK))) != 0;
}

inline void _set(matrix * m, unsigned int y, unsigned int x) {
	int idx = y * m->width + x;
#ifdef DEBUG
	printf("setting   idx %d (%d / %d) for (%d,%d)\n", idx, idx >> LD_S, idx & MASK, y, x);
#endif
	m->data[idx >> LD_S] |= 1 << (idx & MASK);
}

inline void _unset(matrix * m, unsigned int y, unsigned int x) {
	int idx = y * m->width + x;
#ifdef DEBUG
	printf("unsetting idx %d (%d / %d) for (%d,%d): %08x, %08x @ %08x\n", idx, idx >> LD_S, idx & MASK, y, x, m->data[idx >> LD_S], ~((TYPE)1 << (idx & MASK)), (uint)m->data + (idx >> LD_S) * sizeof(TYPE));
#endif
	m->data[idx >> LD_S] &= ~(1 << (idx & MASK));
#ifdef DEBUG
	printf("afterwards: %08x\n", m->data[idx >> LD_S]);
#endif
}

int mtrx_cmp(matrix *m1, matrix *m2) {
	int pwidth = m1->pwidth;
	int dpwidth = pwidth - m2->pwidth;
	if (dpwidth != 0)
		return dpwidth;
	return memcmp(m1->data, m2->data, pwidth);
}

void mtrx_init(matrix **m, const int width, const int height) {
	int padded_width = (width + MASK) >> LD_S;
	*m = (matrix *)malloc(sizeof(matrix));
	matrix *r = *m;
	r->data = (unsigned TYPE *)malloc(height * padded_width * sizeof(TYPE));
	r->width = width;
	r->height = height;
	r->pwidth = padded_width;
}

void _mtrx_init(matrix **m, const int width, const int height) {
	int padded_width = (height * width + MASK) >> LD_S;
	*m = (matrix *)malloc(sizeof(matrix));
	matrix *r = *m;
	r->data = (unsigned TYPE *)malloc(padded_width * sizeof(TYPE));
#ifdef DEBUG
	printf("init: %d, %d, %d @ %08x\n", padded_width, MASK, LD_S, r->data);
#endif
	r->width = width;
	r->height = height;
	r->pwidth = padded_width;
}

void mtrx_clone_data(const matrix *from, matrix *to) {
	memcpy(to->data, from->data, from->pwidth * from->height);
}

void mtrx_fill(matrix *m) {
	if (sizeof(TYPE) == 1) {
		// determine number of full bytes rand() can provide, e.g. 3 for RAND_MAX = 0x7fffffff
		int k = 0xFF;
		int j = 0;
		int b;
		int test = RAND_MAX;
		int max = m->pwidth * m->height;
		
		srand(time(NULL));
		
		for (b = 0; b < sizeof(test) && b < sizeof(int); b++) {
			j |= k << (b << 3);
			if ((test & j) == test) {
				if (test == j)
					b++;
				break;
			}
		}
		
		j = 0;
		while (j < max) {
			int rnd = rand();
			for (k = 0; k < b && j < max; k++, j++)
				m->data[j] = (rnd >> (k << 3)) & 0xFF;
		}
	} else {
		// TODO
	}
}

void mtrx_free(matrix **m) {
	free((*m)->data);
	free(*m);
}
